import base64
import math

from .country import Country
from .color import Color


def total_needed_pixels(level: int) -> int:
    """Return the total number of pixels a user needs to place to reach the
    given `level`.
    """
    return math.ceil(math.pow((level - 1) * math.pow(30, 0.65), (1 / 0.65)))


def needed_pixels(level: int) -> int:
    """Return the number of pixels a user needs to place to reach the next level
    from their current `level`.
    """
    return total_needed_pixels(level + 1) - total_needed_pixels(level)


def decode_color_bitmap(bitmap: int) -> tuple[Color, ...]:
    """Extract list of unlocked premium colors from `extraColorsBitmap` within
    https://backend.wplace.live/me.
    """
    colors = []
    for color in Color:
        if not color.premium:
            continue
        if (bitmap >> (color.id - 32)) & 0b1:
            colors.append(color)
    return tuple(colors)


def decode_country_bitmap(bitmap: str) -> tuple[Country, ...]:
    """Extract list of unlocked country flags from `flagsBitmap` within
    https://backend.wplace.live/me.
    """
    bitmap = int.from_bytes(base64.b64decode(bitmap), byteorder="big")
    countries = []
    for country in Country:
        if (bitmap >> country.id) & 0b1:
            countries.append(country)
    return tuple(countries)



from __future__ import annotations
from pyproj import CRS, Transformer
from rasterio.transform import from_bounds, rowcol, xy
from urllib.parse import urlparse, parse_qs


# setup of the pixel canvas
N_TILES_X: int = 2048
N_TILES_Y: int = 2048
N_TILE_PIXELS_X: int = 1000
N_TILE_PIXELS_Y: int = 1000
N_PIXELS_X: int = N_TILES_X * N_TILE_PIXELS_X
N_PIXELS_Y: int = N_TILES_Y * N_TILE_PIXELS_Y
N_REGION_PIXELS_X: int = 4000
N_REGION_PIXELS_Y: int = 4000
N_REGIONS_X: int = N_PIXELS_X // N_REGION_PIXELS_X
N_REGIONS_Y: int = N_PIXELS_Y // N_REGION_PIXELS_Y

# used coordinate reference systems
EPSG_3857 = CRS.from_epsg(3857) # Pseudo-Mercator
EPSG_4326 = CRS.from_epsg(4326) # longitude / latitude (WGS84)

# transformations between Pseudo-Mercator and WGS84
FROM_LONLAT = Transformer.from_crs(EPSG_4326, EPSG_3857, always_xy=True)
TO_LONLAT = Transformer.from_crs(EPSG_3857, EPSG_4326, always_xy=True)

# WGS84 bounds of the canvas from https://maps.wplace.live/planet
LON_WEST: float = -180.
LON_EAST: float =  180.
LAT_NORTH: float =  85.05113
LAT_SOUTH: float = -85.05113

# canvas bounds in EPSG:3857
WEST, NORTH = FROM_LONLAT.transform(xx=LON_WEST, yy=LAT_NORTH)
EAST, SOUTH = FROM_LONLAT.transform(xx=LON_EAST, yy=LAT_SOUTH)

# transformation between pixel coordinated and EPSG:3857
PIXEL_TRAFO = from_bounds(
    west=WEST,
    north=NORTH,
    east=EAST,
    south=SOUTH,
    width=N_PIXELS_X,
    height=N_PIXELS_Y,
)


def _pixel_to_lonlat(col: int, row: int) -> tuple[float, float]:
    """Return the WGS84 longitude and latitude of a pixel on the Wplace
    canvas.
    """
    x, y = xy(PIXEL_TRAFO, row, col)
    return TO_LONLAT.transform(x, y)


def _lonlat_to_pixel(lon: float, lat: float) -> tuple[int, int]:
    """Return the pixel coordinate on the Wplace canvas corresponding to the
    given WGS84 longitude and latitude.
    """
    x, y = FROM_LONLAT.transform(lon, lat)
    row, col = rowcol(PIXEL_TRAFO, x, y)
    return col.item(), row.item()


class _CanvasElement(tuple):
    __slots__ = ()

    # number of elements in x and y direction
    N_X: int = None
    N_Y: int = None

    def __new__(cls, x: int, y: int):
        if not (isinstance(x, int) and isinstance(y, int)):
            raise TypeError("x and y must be integers")
        if cls.N_X is None or cls.N_Y is None:
            raise NotImplementedError(
                f"{cls.__name__} must define class attributes N_X and N_Y")
        if not (0 <= x < cls.N_X and 0 <= y < cls.N_Y):
            raise ValueError(
                f"x must be in [0,{cls.N_X}), y must be in [0,{cls.N_Y})")
        return super().__new__(cls, (x, y))

    @property
    def x(self) -> int:
        return self[0]

    @property
    def y(self) -> int:
        return self[1]

    def __str__(self) -> str:
        return f"({self.x}, {self.y})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.x}, y={self.y})"

    def __add__(self, other: tuple[int, int]) -> _CanvasElement:
        x, y = other
        return type(self)(
            (self.x + x) % self.N_X,
            self.y + y,
        )

    def __sub__(
        self,
        other: tuple[int, int] | _CanvasElement,
    ) -> _CanvasElement | tuple[int, int]:
        if isinstance(other, _CanvasElement):
            dx = self.x - other.x
            if dx < 0:
                dx += self.N_X
            return dx, self.y - other.y
        x, y = other
        return type(self)(
            (self.x - x) % self.N_X,
            self.y - y,
        )



class Region(_CanvasElement):
    """A 4000 x 4000 pixels region on the Wplace canvas."""
    N_X: int = N_REGIONS_X
    N_Y: int = N_REGIONS_Y

    @property
    def origin(self) -> Pixel:
        """Return the most north-west pixel of the region."""
        return Pixel(
            self.x * N_REGION_PIXELS_X,
            self.y * N_REGION_PIXELS_Y,
        )


class Tile(_CanvasElement):
    """A 1000 x 1000 pixels tile on the Wplace canvas."""
    N_X: int = N_TILES_X
    N_Y: int = N_TILES_Y

    @property
    def origin(self) -> Pixel:
        """Return the most north-west pixel of the tile."""
        return Pixel(
            self.x * N_TILE_PIXELS_X,
            self.y * N_TILE_PIXELS_Y,
        )

    @property
    def region(self) -> Region:
        """Return the canvas region that contains the tile."""
        return self.origin.region

class Pixel(_CanvasElement):
    """A pixel on the Wplace canvas."""
    N_X: int = N_PIXELS_X
    N_Y: int = N_PIXELS_Y

    @classmethod
    def from_tile(
        cls,
        tile: Tile,
        px: int = 0,
        py: int = 0,
    ) -> Pixel:
        if not 0 <= px < N_TILE_PIXELS_X:
            raise ValueError(f"px has to be in [{0},{N_TILE_PIXELS_X})")
        if not 0 <= py < N_TILE_PIXELS_Y:
            raise ValueError(f"py has to be in [{0},{N_TILE_PIXELS_Y})")
        return tile.origin + (px, py)

    @classmethod
    def from_region(
        cls,
        region: Region,
        px: int = 0,
        py: int = 0,
    ) -> Pixel:
        if not 0 <= px < N_REGION_PIXELS_X:
            raise ValueError(f"px has to be in [{0},{N_REGION_PIXELS_X})")
        if not 0 <= py < N_REGION_PIXELS_Y:
            raise ValueError(f"py has to be in [{0},{N_REGION_PIXELS_Y})")
        return region.origin + (px, py)

    @classmethod
    def from_lonlat(cls, lon: float, lat: float) -> Pixel:
        x, y = _lonlat_to_pixel(lon=lon, lat=lat)
        return cls(x, y)

    @classmethod
    def from_link(cls, url: str) -> Pixel:
        parsed = urlparse(url)
        params = parse_qs(parsed.query)
        return cls.from_lonlat(
            lon=float(params["lng"][0]),
            lat=float(params["lat"][0]),
        )

    @property
    def tile(self) -> Tile:
        """Return the canvas tile that contains the pixel."""
        return Tile(
            self.x // N_TILE_PIXELS_X,
            self.y // N_TILE_PIXELS_Y,
        )

    @property
    def coords_in_tile(self) -> tuple[int, int]:
        """Return coordinates of the pixel within its canvas tile."""
        return (
            self.x % N_TILE_PIXELS_X,
            self.y % N_TILE_PIXELS_Y,
        )

    @property
    def region(self) -> Region:
        """Return the canvas region that contains the pixel."""
        return Region(
            self.x // N_REGION_PIXELS_X,
            self.y // N_REGION_PIXELS_Y,
        )

    @property
    def coords_in_region(self) -> tuple[int, int]:
        """Return coordinates of the pixel within its canvas region."""
        return (
            self.x % N_REGION_PIXELS_X,
            self.y % N_REGION_PIXELS_Y,
        )

    @property
    def lonlat(self) -> tuple[float, float]:
        """Return the WGS84 longitude and latitude of the pixel."""
        return _pixel_to_lonlat(col=self.x, row=self.y)



import cloudscraper
from io import BytesIO
import logging
from PIL import Image
import requests

from .canvas import Pixel, Region, Tile
from .dto import PixelInfo, RegionInfo


class WplaceAPI:
    def __init__(
        self,
        frontend_url: str = "https://wplace.live",
        backend_url: str = "https://backend.wplace.live",
    ) -> None:
        self.frontend_url = frontend_url.rstrip("/")
        self.backend_url = backend_url.rstrip("/")

    def get_pixel_link(
        self,
        pixel: Pixel,
        select: bool = False,
        zoom: float | None = None,
    ) -> str:
        """Return link that navigates to the `pixel` on the Wplace canvas."""
        lon, lat = pixel.lonlat
        url = f"{self.frontend_url}/?lat={lat}&lng={lon}"
        if zoom is not None:
            url += f"&zoom={zoom}"
        if select:
            url += "&select=0"
        return url

    def get_pixel_url(self, pixel: Pixel) -> str:
        """Return the backend URL used to get information about the `pixel`."""
        tx, ty = pixel.tile
        px, py = pixel.coords_in_tile
        return f"{self.backend_url}/s0/pixel/{tx}/{ty}?x={px}&y={py}"

    def get_tile_url(self, tile: Tile) -> str:
        """Return the backend URL of the `tile` image."""
        return f"{self.backend_url}/files/s0/tiles/{tile.x}/{tile.y}.png"

    def fetch_pixel_info(self, pixel: Pixel) -> PixelInfo:
        """Get information about the `pixel` from the backend."""
        # TODO: Handle exceptions / add retries + timeout
        url = self.get_pixel_url(pixel=pixel)
        logging.debug(f"Downloading pixel information from {url}")
        scraper = cloudscraper.create_scraper()
        response = scraper.get(url)
        info = response.json()
        del scraper

        return PixelInfo.from_dict(info=info)

    def fetch_region_info(self, region: Region) -> RegionInfo:
        """Get information about the `region` from the backend."""
        return self.fetch_pixel_info(region.origin).region

    def fetch_tile_image(self, tile: Tile) -> Image.Image | None:
        """Download the `tile` image from the backend.

        Args:
            tile: The tile to download the image for.

        Returns:
            Image.Image: The tile image. Will be an indexed image with an RGB
                color palette with one palette index assigned to full
                transparency. NOTE: Palette length, order, and transparency
                index are NOT consistent across different tile images!
            None: If errors occurred or no image exists for the tile, yet.
        """
        url = self.get_tile_url(tile=tile)
        # TODO: Handle exceptions / add retries + timeout
        logging.debug(f"Downloading tile image from {url}")
        try:
            response = requests.get(url)
            response.raise_for_status()
            image = Image.open(BytesIO(response.content))
            return image
        except requests.HTTPError as exc:
            logging.error(f"Failed to load tile image: {exc}")
            return None



from __future__ import annotations
from dataclasses import dataclass

from .country import Country, from_id


@dataclass(order=True)
class UserInfo:
    id: int
    name: str
    alliance_id: int | None
    alliance_name: str | None
    country: Country | None = None
    discord_name: str | None = None
    #TODO: picture: Image.Image | None = None

    @classmethod
    def from_dict(cls, info: dict[str, str | int]) -> UserInfo:
        alliance_id = info["allianceId"]
        if alliance_id == 0:
            alliance_id = None

        alliance_name = info["allianceName"]
        if alliance_name == "":
            alliance_name = None

        country_id = info["equippedFlag"]
        if country_id == 0:
            country = None
        else:
            country = from_id(country_id)

        return cls(
            id=info["id"],
            name=info["name"],
            alliance_id=alliance_id,
            alliance_name=alliance_name,
            country=country,
            discord_name=info.get("discord", None),
        )


@dataclass(frozen=True, order=True)
class RegionInfo:
    id: int
    city_id: int
    name: str
    number: int
    country: Country | None = None

    def __str__(self) -> str:
        return f"{self.name} #{self.number}"

    @classmethod
    def from_dict(cls, info: dict[str, str | int]) -> RegionInfo:
        country_id = info["countryId"]
        if country_id == 0:
            country = None
        else:
            country = from_id(country_id)

        return cls(
            id=info["id"],
            city_id=info["cityId"],
            name=info["name"],
            number=info["number"],
            country=country,
        )


@dataclass
class PixelInfo:
    painted_by: UserInfo | None
    region: RegionInfo

    @classmethod
    def from_dict(cls, info: dict[str, dict[str, str | int]]) -> PixelInfo:
        user = UserInfo.from_dict(info["paintedBy"])
        if user.id == 0:
            user = None

        region = RegionInfo.from_dict(info["region"])

        return cls(painted_by=user, region=region)



from __future__ import annotations
from dataclasses import dataclass
import enum


def _rgb_to_hex(red: int, green: int, blue: int, bit_depth: int = 8) -> int:
    hex = 0
    for i, channel in enumerate((blue, green, red)):
        hex |= channel << (i * bit_depth)
    return hex


def _hex_to_rgb(hex: int | str, bit_depth: int = 8) -> tuple[int, int, int]:
    if isinstance(hex, str):
        hex = int(hex, 16)
    mask = (1 << bit_depth) - 1
    return tuple([hex >> (i * bit_depth) & mask for i in range(2, -1, -1)])


@dataclass(frozen=True, order=True)
class _Color:
    id: int
    label: str
    hex: int
    premium: bool

    @property
    def rgb(self) -> tuple[int, int, int]:
        return _hex_to_rgb(self.hex)

    def __str__(self) -> str:
        return self.label


class Color(_Color, enum.Enum):
    BLACK =             1, "Black",            0x000000, False
    DARK_GRAY =         2, "Dark Gray",        0x3c3c3c, False
    GRAY =              3, "Gray",             0x787878, False
    MEDIUM_GRAY =      32, "Medium Gray",      0xaaaaaa, True
    LIGHT_GRAY =        4, "Light Gray",       0xd2d2d2, False
    WHITE =             5, "White",            0xffffff, False
    DEEP_RED =          6, "Deep Red",         0x600018, False
    DARK_RED =         33, "Dark Red",         0xa50e1e, True
    RED =               7, "Red",              0xed1c24, False
    LIGHT_RED =        34, "Light Red",        0xfa8072, True
    DARK_ORANGE =      35, "Dark Orange",      0xe45c1a, True
    ORANGE =            8, "Orange",           0xff7f27, False
    GOLD =              9, "Gold",             0xf6aa09, False
    YELLOW =           10, "Yellow",           0xf9dd3b, False
    LIGHT_YELLOW =     11, "Light Yellow",     0xfffabc, False
    DARK_GOLDENROD =   37, "Dark Goldenrod",   0x9c8431, True
    GOLDENROD =        38, "Goldenrod",        0xc5ad31, True
    LIGHT_GOLDENROD =  39, "Light Goldenrod",  0xe8d45f, True
    DARK_OLIVE =       40, "Dark Olive",       0x4a6b3a, True
    OLIVE =            41, "Olive",            0x5a944a, True
    LIGHT_OLIVE =      42, "Light Olive",      0x84c573, True
    DARK_GREEN =       12, "Dark Green",       0x0eb968, False
    GREEN =            13, "Green",            0x13e67b, False
    LIGHT_GREEN =      14, "Light Green",      0x87ff5e, False
    DARK_TEAL =        15, "Dark Teal",        0x0c816e, False
    TEAL =             16, "Teal",             0x10aea6, False
    LIGHT_TEAL =       17, "Light Teal",       0x13e1be, False
    DARK_CYAN =        43, "Dark Cyan",        0x0f799f, True
    CYAN =             20, "Cyan",             0x60f7f2, False
    LIGHT_CYAN =       44, "Light Cyan",       0xbbfaf2, True
    DARK_BLUE =        18, "Dark Blue",        0x28509e, False
    BLUE =             19, "Blue",             0x4093e4, False
    LIGHT_BLUE =       45, "Light Blue",       0x7dc7ff, True
    DARK_INDIGO =      46, "Dark Indigo",      0x4d31b8, True
    INDIGO =           21, "Indigo",           0x6b50f6, False
    LIGHT_INDIGO =     22, "Light Indigo",     0x99b1fb, False
    DARK_SLATE_BLUE =  47, "Dark Slate Blue",  0x4a4284, True
    SLATE_BLUE =       48, "Slate Blue",       0x7a71c4, True
    LIGHT_SLATE_BLUE = 49, "Light Slate Blue", 0xb5aef1, True
    DARK_PURPLE =      23, "Dark Purple",      0x780c99, False
    PURPLE =           24, "Purple",           0xaa38b9, False
    LIGHT_PURPLE =     25, "Light Purple",     0xe09ff9, False
    DARK_PINK =        26, "Dark Pink",        0xcb007a, False
    PINK =             27, "Pink",             0xec1f80, False
    LIGHT_PINK =       28, "Light Pink",       0xf38da9, False
    DARK_PEACH =       53, "Dark Peach",       0x9b5249, True
    PEACH =            54, "Peach",            0xd18078, True
    LIGHT_PEACH =      55, "Light Peach",      0xfab6a4, True
    DARK_BROWN =       29, "Dark Brown",       0x684634, False
    BROWN =            30, "Brown",            0x95682a, False
    LIGHT_BROWN =      50, "Light Brown",      0xdba463, True
    DARK_TAN =         56, "Dark Tan",         0x7b6352, True
    TAN =              57, "Tan",              0x9c846b, True
    LIGHT_TAN =        36, "Light Tan",        0xd6b594, True
    DARK_BEIGE =       51, "Dark Beige",       0xd18051, True
    BEIGE =            31, "Beige",            0xf8b277, False
    LIGHT_BEIGE =      52, "Light Beige",      0xffc5a5, True
    DARK_STONE =       61, "Dark Stone",       0x6d643f, True
    STONE =            62, "Stone",            0x948c6b, True
    LIGHT_STONE =      63, "Light Stone",      0xcdc59e, True
    DARK_SLATE =       58, "Dark Slate",       0x333941, True
    SLATE =            59, "Slate",            0x6d758d, True
    LIGHT_SLATE =      60, "Light Slate",      0xb3b9d1, True
    TRANSPARENT =       0, "Transparent",      0xdeface, False


_ID_MAP: dict[int, Color] = {c.id: c for c in Color}
_HEX_MAP: dict[int, Color] = {c.hex: c for c in Color}
_RGB_MAP: dict[tuple[int, int, int], Color] = {c.rgb: c for c in Color}


def from_id(id: int) -> Color:
    return _ID_MAP[id]


def from_hex(hex: int | str) -> Color:
    if isinstance(hex, str):
        hex = int(hex, 16)
    return _HEX_MAP[hex]


def from_rgb(rgb: tuple[int, int, int]) -> Color:
    return _RGB_MAP[rgb]



from dataclasses import dataclass
import enum


def _flag_to_iso(flag: str) -> str:
    return "".join(
        chr(ord(c) - 0x1F1E6 + ord("A"))
        for c in flag
    )


def _iso_to_flag(iso: str) -> str:
    return "".join(
        chr(ord("🇦") + ord(c) - ord("A"))
        for c in iso.upper()
    )


@dataclass(frozen=True, order=True)
class _Country:
    id: int
    iso: str
    label: str

    @property
    def flag(self) -> str:
        return _iso_to_flag(self.iso)

    def __str__(self) -> str:
        return self.label


class Country(_Country, enum.Enum):
    AF =   1, "AF", "Afghanistan"
    AL =   2, "AL", "Albania"
    DZ =   3, "DZ", "Algeria"
    AS =   4, "AS", "American Samoa"
    AD =   5, "AD", "Andorra"
    AO =   6, "AO", "Angola"
    AI =   7, "AI", "Anguilla"
    AQ =   8, "AQ", "Antarctica"
    AG =   9, "AG", "Antigua and Barbuda"
    AR =  10, "AR", "Argentina"
    AM =  11, "AM", "Armenia"
    AW =  12, "AW", "Aruba"
    AU =  13, "AU", "Australia"
    AT =  14, "AT", "Austria"
    AZ =  15, "AZ", "Azerbaijan"
    BS =  16, "BS", "Bahamas"
    BH =  17, "BH", "Bahrain"
    BD =  18, "BD", "Bangladesh"
    BB =  19, "BB", "Barbados"
    BY =  20, "BY", "Belarus"
    BE =  21, "BE", "Belgium"
    BZ =  22, "BZ", "Belize"
    BJ =  23, "BJ", "Benin"
    BM =  24, "BM", "Bermuda"
    BT =  25, "BT", "Bhutan"
    BO =  26, "BO", "Bolivia"
    BQ =  27, "BQ", "Bonaire"
    BA =  28, "BA", "Bosnia and Herzegovina"
    BW =  29, "BW", "Botswana"
    BV =  30, "BV", "Bouvet Island"
    BR =  31, "BR", "Brazil"
    IO =  32, "IO", "British Indian Ocean Territory"
    BN =  33, "BN", "Brunei Darussalam"
    BG =  34, "BG", "Bulgaria"
    BF =  35, "BF", "Burkina Faso"
    BI =  36, "BI", "Burundi"
    CV =  37, "CV", "Cabo Verde"
    KH =  38, "KH", "Cambodia"
    CM =  39, "CM", "Cameroon"
    CA =  40, "CA", "Canada"
    KY =  41, "KY", "Cayman Islands"
    CF =  42, "CF", "Central African Republic"
    TD =  43, "TD", "Chad"
    CL =  44, "CL", "Chile"
    CN =  45, "CN", "China"
    CX =  46, "CX", "Christmas Island"
    CC =  47, "CC", "Cocos (Keeling) Islands"
    CO =  48, "CO", "Colombia"
    KM =  49, "KM", "Comoros"
    CG =  50, "CG", "Congo"
    CK =  51, "CK", "Cook Islands"
    CR =  52, "CR", "Costa Rica"
    HR =  53, "HR", "Croatia"
    CU =  54, "CU", "Cuba"
    CW =  55, "CW", "Curaçao"
    CY =  56, "CY", "Cyprus"
    CZ =  57, "CZ", "Czechia"
    CI =  58, "CI", "Côte d'Ivoire"
    DK =  59, "DK", "Denmark"
    DJ =  60, "DJ", "Djibouti"
    DM =  61, "DM", "Dominica"
    DO =  62, "DO", "Dominican Republic"
    EC =  63, "EC", "Ecuador"
    EG =  64, "EG", "Egypt"
    SV =  65, "SV", "El Salvador"
    GQ =  66, "GQ", "Equatorial Guinea"
    ER =  67, "ER", "Eritrea"
    EE =  68, "EE", "Estonia"
    SZ =  69, "SZ", "Eswatini"
    ET =  70, "ET", "Ethiopia"
    FK =  71, "FK", "Falkland Islands (Malvinas)"
    FO =  72, "FO", "Faroe Islands"
    FJ =  73, "FJ", "Fiji"
    FI =  74, "FI", "Finland"
    FR =  75, "FR", "France"
    GF =  76, "GF", "French Guiana"
    PF =  77, "PF", "French Polynesia"
    TF =  78, "TF", "French Southern Territories"
    GA =  79, "GA", "Gabon"
    GM =  80, "GM", "Gambia"
    GE =  81, "GE", "Georgia"
    DE =  82, "DE", "Germany"
    GH =  83, "GH", "Ghana"
    GI =  84, "GI", "Gibraltar"
    GR =  85, "GR", "Greece"
    GL =  86, "GL", "Greenland"
    GD =  87, "GD", "Grenada"
    GP =  88, "GP", "Guadeloupe"
    GU =  89, "GU", "Guam"
    GT =  90, "GT", "Guatemala"
    GG =  91, "GG", "Guernsey"
    GN =  92, "GN", "Guinea"
    GW =  93, "GW", "Guinea-Bissau"
    GY =  94, "GY", "Guyana"
    HT =  95, "HT", "Haiti"
    HM =  96, "HM", "Heard Island and McDonald Islands"
    HN =  97, "HN", "Honduras"
    HK =  98, "HK", "Hong Kong"
    HU =  99, "HU", "Hungary"
    IS = 100, "IS", "Iceland"
    IN = 101, "IN", "India"
    ID = 102, "ID", "Indonesia"
    IR = 103, "IR", "Iran"
    IQ = 104, "IQ", "Iraq"
    IE = 105, "IE", "Ireland"
    IM = 106, "IM", "Isle of Man"
    IL = 107, "IL", "Israel"
    IT = 108, "IT", "Italy"
    JM = 109, "JM", "Jamaica"
    JP = 110, "JP", "Japan"
    JE = 111, "JE", "Jersey"
    JO = 112, "JO", "Jordan"
    KZ = 113, "KZ", "Kazakhstan"
    KE = 114, "KE", "Kenya"
    KI = 115, "KI", "Kiribati"
    XK = 116, "XK", "Kosovo"
    KW = 117, "KW", "Kuwait"
    KG = 118, "KG", "Kyrgyzstan"
    LA = 119, "LA", "Laos"
    LV = 120, "LV", "Latvia"
    LB = 121, "LB", "Lebanon"
    LS = 122, "LS", "Lesotho"
    LR = 123, "LR", "Liberia"
    LY = 124, "LY", "Libya"
    LI = 125, "LI", "Liechtenstein"
    LT = 126, "LT", "Lithuania"
    LU = 127, "LU", "Luxembourg"
    MO = 128, "MO", "Macao"
    MG = 129, "MG", "Madagascar"
    MW = 130, "MW", "Malawi"
    MY = 131, "MY", "Malaysia"
    MV = 132, "MV", "Maldives"
    ML = 133, "ML", "Mali"
    MT = 134, "MT", "Malta"
    MH = 135, "MH", "Marshall Islands"
    MQ = 136, "MQ", "Martinique"
    MR = 137, "MR", "Mauritania"
    MU = 138, "MU", "Mauritius"
    YT = 139, "YT", "Mayotte"
    MX = 140, "MX", "Mexico"
    FM = 141, "FM", "Micronesia"
    MD = 142, "MD", "Moldova"
    MC = 143, "MC", "Monaco"
    MN = 144, "MN", "Mongolia"
    ME = 145, "ME", "Montenegro"
    MS = 146, "MS", "Montserrat"
    MA = 147, "MA", "Morocco"
    MZ = 148, "MZ", "Mozambique"
    MM = 149, "MM", "Myanmar"
    NA = 150, "NA", "Namibia"
    NR = 151, "NR", "Nauru"
    NP = 152, "NP", "Nepal"
    NL = 153, "NL", "Netherlands"
    NC = 154, "NC", "New Caledonia"
    NZ = 155, "NZ", "New Zealand"
    NI = 156, "NI", "Nicaragua"
    NE = 157, "NE", "Niger"
    NG = 158, "NG", "Nigeria"
    NU = 159, "NU", "Niue"
    NF = 160, "NF", "Norfolk Island"
    KP = 161, "KP", "North Korea"
    MK = 162, "MK", "North Macedonia"
    MP = 163, "MP", "Northern Mariana Islands"
    NO = 164, "NO", "Norway"
    OM = 165, "OM", "Oman"
    PK = 166, "PK", "Pakistan"
    PW = 167, "PW", "Palau"
    PS = 168, "PS", "Palestine"
    PA = 169, "PA", "Panama"
    PG = 170, "PG", "Papua New Guinea"
    PY = 171, "PY", "Paraguay"
    PE = 172, "PE", "Peru"
    PH = 173, "PH", "Philippines"
    PN = 174, "PN", "Pitcairn"
    PL = 175, "PL", "Poland"
    PT = 176, "PT", "Portugal"
    PR = 177, "PR", "Puerto Rico"
    QA = 178, "QA", "Qatar"
    CD = 179, "CD", "Republic of the Congo"
    RO = 180, "RO", "Romania"
    RU = 181, "RU", "Russia"
    RW = 182, "RW", "Rwanda"
    RE = 183, "RE", "Réunion"
    BL = 184, "BL", "Saint Barthélemy"
    SH = 185, "SH", "Saint Helena"
    KN = 186, "KN", "Saint Kitts and Nevis"
    LC = 187, "LC", "Saint Lucia"
    MF = 188, "MF", "Saint Martin (French part)"
    PM = 189, "PM", "Saint Pierre and Miquelon"
    VC = 190, "VC", "Saint Vincent and the Grenadines"
    WS = 191, "WS", "Samoa"
    SM = 192, "SM", "San Marino"
    ST = 193, "ST", "Sao Tome and Principe"
    SA = 194, "SA", "Saudi Arabia"
    SN = 195, "SN", "Senegal"
    RS = 196, "RS", "Serbia"
    SC = 197, "SC", "Seychelles"
    SL = 198, "SL", "Sierra Leone"
    SG = 199, "SG", "Singapore"
    SX = 200, "SX", "Sint Maarten (Dutch part)"
    SK = 201, "SK", "Slovakia"
    SI = 202, "SI", "Slovenia"
    SB = 203, "SB", "Solomon Islands"
    SO = 204, "SO", "Somalia"
    ZA = 205, "ZA", "South Africa"
    GS = 206, "GS", "South Georgia and the South Sandwich Islands"
    KR = 207, "KR", "South Korea"
    SS = 208, "SS", "South Sudan"
    ES = 209, "ES", "Spain"
    LK = 210, "LK", "Sri Lanka"
    SD = 211, "SD", "Sudan"
    SR = 212, "SR", "Suriname"
    SJ = 213, "SJ", "Svalbard and Jan Mayen"
    SE = 214, "SE", "Sweden"
    CH = 215, "CH", "Switzerland"
    SY = 216, "SY", "Syrian Arab Republic"
    TW = 217, "TW", "Taiwan"
    TJ = 218, "TJ", "Tajikistan"
    TZ = 219, "TZ", "Tanzania"
    TH = 220, "TH", "Thailand"
    TL = 221, "TL", "Timor-Leste"
    TG = 222, "TG", "Togo"
    TK = 223, "TK", "Tokelau"
    TO = 224, "TO", "Tonga"
    TT = 225, "TT", "Trinidad and Tobago"
    TN = 226, "TN", "Tunisia"
    TM = 227, "TM", "Turkmenistan"
    TC = 228, "TC", "Turks and Caicos Islands"
    TV = 229, "TV", "Tuvalu"
    TR = 230, "TR", "Türkiye"
    UG = 231, "UG", "Uganda"
    UA = 232, "UA", "Ukraine"
    AE = 233, "AE", "United Arab Emirates"
    GB = 234, "GB", "United Kingdom"
    US = 235, "US", "United States"
    UM = 236, "UM", "United States Minor Outlying Islands"
    UY = 237, "UY", "Uruguay"
    UZ = 238, "UZ", "Uzbekistan"
    VU = 239, "VU", "Vanuatu"
    VA = 240, "VA", "Vatican City"
    VE = 241, "VE", "Venezuela"
    VN = 242, "VN", "Viet Nam"
    VG = 243, "VG", "Virgin Islands"
    VI = 244, "VI", "Virgin Islands"
    WF = 245, "WF", "Wallis and Futuna"
    EH = 246, "EH", "Western Sahara"
    YE = 247, "YE", "Yemen"
    ZM = 248, "ZM", "Zambia"
    ZW = 249, "ZW", "Zimbabwe"
    AX = 250, "AX", "Åland Islands"
    IC = 251, "IC", "Canary Islands"


_ID_MAP: dict[int, Country] = {c.id: c for c in Country}
_ISO_MAP: dict[str, Country] = {c.iso: c for c in Country}
_FLAG_MAP: dict[str, Country] = {c.flag: c for c in Country}


def from_id(id: int) -> Country:
    return _ID_MAP[id]


def from_iso(iso: str) -> Country:
    return _ISO_MAP[iso]


def from_flag(flag: str) -> Country:
    return _FLAG_MAP[flag]



import itertools
import numpy as np
from PIL import Image
import time

from .canvas import (
    N_TILE_PIXELS_X,
    N_TILE_PIXELS_X,
    N_TILE_PIXELS_Y,
    N_TILE_PIXELS_Y,
    Pixel,
)
from .country import Country
from .color import Color
# from .image_processing import standardize_color_palette


COLOR_ICON_BITS: dict[Color, int] = {
    Color.BLACK:             4897444,
    Color.DARK_GRAY:         4756004,
    Color.GRAY:             15241774,
    Color.MEDIUM_GRAY:      31850982,
    Color.LIGHT_GRAY:       11065002,
    Color.WHITE:            15269550,
    Color.DEEP_RED:         33209205,
    Color.DARK_RED:         19267878,
    Color.RED:              15728622,
    Color.LIGHT_RED:        16236308,
    Color.DARK_ORANGE:      33481548,
    Color.ORANGE:           15658734,
    Color.GOLD:             33226431,
    Color.YELLOW:           33391295,
    Color.LIGHT_YELLOW:     32641727,
    Color.DARK_GOLDENROD:   22708917,
    Color.GOLDENROD:        14352822,
    Color.LIGHT_GOLDENROD:   7847326,
    Color.DARK_OLIVE:        7652956,
    Color.OLIVE:            22501038,
    Color.LIGHT_OLIVE:      28457653,
    Color.DARK_GREEN:       15589098,
    Color.GREEN:            11516906,
    Color.LIGHT_GREEN:       9760338,
    Color.DARK_TEAL:        15399560,
    Color.TEAL:              4685802,
    Color.LIGHT_TEAL:       15587182,
    Color.DARK_CYAN:         9179234,
    Color.CYAN:             29206876,
    Color.LIGHT_CYAN:       30349539,
    Color.DARK_BLUE:         3570904,
    Color.BLUE:             15259182,
    Color.LIGHT_BLUE:        4685269,
    Color.DARK_INDIGO:      18295249,
    Color.INDIGO:           29224831,
    Color.LIGHT_INDIGO:     21427311,
    Color.DARK_SLATE_BLUE:  26843769,
    Color.SLATE_BLUE:       24483191,
    Color.LIGHT_SLATE_BLUE:  5211003,
    Color.DARK_PURPLE:      22511061,
    Color.PURPLE:           15161013,
    Color.LIGHT_PURPLE:      4667844,
    Color.DARK_PINK:        11392452,
    Color.PINK:             11375466,
    Color.LIGHT_PINK:        6812424,
    Color.DARK_PEACH:       14829567,
    Color.PEACH:            17971345,
    Color.LIGHT_PEACH:      28873275,
    Color.DARK_BROWN:        5225454,
    Color.BROWN:            29197179,
    Color.LIGHT_BROWN:       4681156,
    Color.DARK_TAN:         21392581,
    Color.TAN:               7460636,
    Color.LIGHT_TAN:        23013877,
    Color.DARK_BEIGE:       29010254,
    Color.BEIGE:            18285009,
    Color.LIGHT_BEIGE:      18846257,
    Color.DARK_STONE:       21825364,
    Color.STONE:            29017787,
    Color.LIGHT_STONE:       4357252,
    Color.DARK_SLATE:       23057550,
    Color.SLATE:            26880179,
    Color.LIGHT_SLATE:       5242308,
    Color.TRANSPARENT:      15237450,
}


def get_rgb_color_map() -> dict[tuple[int, int, int], Color]:
    return {color.rgb: color for color in Color}


def get_rgb_palette(
    sort_by: str = "id",
    incl_trans: bool = False,
    incl_alpha: bool = False,
    flatten: bool = False,
) -> list[tuple[int, ...]] | list[int]:
    palette = []

    if sort_by == "palette":
        sorter = lambda color: tuple(Color).index(color)
    else:
        sorter = lambda color: getattr(color, sort_by)

    for color in sorted(Color, key=sorter):
        if color == Color.TRANSPARENT:
            alpha = 0
            if not incl_trans:
                continue
        else:
            alpha = 255

        color_tuple = color.rgb
        if incl_alpha:
            color_tuple = tuple(list(color_tuple) + [alpha])

        if flatten:
            palette.extend(color_tuple)
        else:
            palette.append(color_tuple)

    return palette


def get_color_icon(color: Color) -> Image.Image:
    n_pixels: int = 5
    bitmap = COLOR_ICON_BITS[color]
    bits = [bitmap >> i & 0b1 for i in range(n_pixels * n_pixels)]
    array = np.array(bits, dtype=np.uint8).reshape((n_pixels, n_pixels))
    image = Image.fromarray(array)
    image.putpalette([0, 0, 0] + list(color.rgb))
    image.info["transparency"] = 0
    return image


def download_area_image(
    origin: Pixel,
    width: int,
    height: int,
    sleep_time: float = 1.,
) -> Image.Image:
    """Load all needed tile images from the backend and create an image of the
    specified area.

    Returns:
        Image.Image: Indexed image with the Wplace color palette (63 solid
            colors + full transparency).
    """
    # TODO: Implement incremental download / save to handle large areas
    # TODO: Implement option to use already downloaded tile images
    start_tile = origin.tile
    x_start, y_start = origin.tile_pixel

    n_cols = 1 + (x_start + width - 1) // canvas.N_TILE_PIXELS_X
    n_rows = 1 + (y_start + height - 1) // canvas.N_TILE_PIXELS_Y
    n_tiles = n_cols * n_rows

    # init empty array to stitch tiles
    trans_index = Color.TRANSPARENT.id
    array_width = n_cols * canvas.N_TILE_PIXELS_X
    array_height = n_rows * canvas.N_TILE_PIXELS_Y
    array = np.ones((array_height, array_width), dtype=np.uint8) * trans_index

    for i, (col, row) in enumerate(itertools.product(range(n_cols), range(n_rows))):
        try:
            tile = start_tile + (col, row)
        except ValueError:
            # TODO: Handle out-of-bounds better
            continue
        tile_image = tile.download_image(save=False)
        if tile_image is None:
            continue
        tile_image = standardize_color_palette(tile_image)
        x0 = col * canvas.N_TILE_PIXELS_X
        x1 = x0 + canvas.N_TILE_PIXELS_X
        y0 = row * canvas.N_TILE_PIXELS_Y
        y1 = y0 + canvas.N_TILE_PIXELS_Y
        array[y0:y1, x0:x1] = np.array(tile_image, dtype=np.uint8)
        del tile_image
        if i < n_tiles - 1:
            time.sleep(sleep_time)

    cutout = array[y_start:y_start + height, x_start:x_start + width]
    image = Image.fromarray(cutout)
    image.putpalette(Color.rgb_palette(incl_trans=True, flatten=True))
    image.info["transparency"] = trans_index
    return image
